from typing import Optional

import pygame
from pygame.constants import HWSURFACE, DOUBLEBUF, RESIZABLE
from pygame.surface import Surface

from .MouseButtons import MouseButtons
from .MouseWheel import MouseWheel
from .Theme import Theme
from .mode.GameMode import GameMode


class UserInterface:
    def __init__(self, theme: Theme):

        # Create window with default resolution
        pygame.init()
        self.__window = pygame.display.set_mode((1024, 768), HWSURFACE | DOUBLEBUF | RESIZABLE)
        pygame.display.set_caption("2D Medieval Strategy Game with Python, http://www.patternsgameprog.com")
        pygame.display.set_icon(pygame.image.load("assets/toen/icon.png"))

        # Rendering
        self.__theme = theme
        self.__rescaledX = 0
        self.__rescaledY = 0
        self.__rescaledScaleX = 1.0
        self.__rescaledScaleY = 1.0
        self.__renderWidth = self.__window.get_width()
        self.__renderHeight = self.__window.get_height()

        # Inputs
        self.__mouseFocus = False

        # Other
        self.__gameMode: Optional[GameMode] = None
        self.__running = True
        self.__clock = pygame.time.Clock()

    @property
    def theme(self) -> Theme:
        return self.__theme

    def setGameMode(self, gameMode: GameMode):
        self.__gameMode = gameMode

    def setRenderSize(self, renderWidth: int, renderHeight: int):
        self.__renderWidth = renderWidth
        self.__renderHeight = renderHeight

    def __processKeyEvent(self, event):
        if self.__gameMode is None:
            return

        if event.type == pygame.KEYDOWN:
            self.__gameMode.keyDown(event.key)
        elif event.type == pygame.KEYUP:
            self.__gameMode.keyUp(event.key)

    def __processMouseEvent(self, event):
        if self.__gameMode is None:
            return

        if event.type == pygame.ACTIVEEVENT:
            if self.__mouseFocus:
                self.__mouseFocus = False
                self.__gameMode.mouseLeave()
            return

        mouseX, mouseY = pygame.mouse.get_pos()
        mouseX = int((mouseX - self.__rescaledX) / self.__rescaledScaleX)
        mouseY = int((mouseY - self.__rescaledY) / self.__rescaledScaleY)
        pygameButtons = pygame.mouse.get_pressed(num_buttons=3)
        buttons = MouseButtons(pygameButtons[0], pygameButtons[1], pygameButtons[2])
        if 0 <= mouseX < self.__renderWidth \
                and 0 <= mouseY < self.__renderHeight:
            if not self.__mouseFocus:
                self.__mouseFocus = True
                self.__gameMode.mouseEnter(mouseX, mouseY, buttons)
            if event.type == pygame.MOUSEBUTTONDOWN:
                self.__gameMode.mouseButtonDown(mouseX, mouseY, buttons)
            elif event.type == pygame.MOUSEBUTTONUP:
                self.__gameMode.mouseButtonUp(mouseX, mouseY, buttons)
            elif event.type == pygame.MOUSEWHEEL:
                wheel = MouseWheel(event.x, event.y, event.flipped, event.which)
                self.__gameMode.mouseWheel(mouseX, mouseY, buttons, wheel)
            elif event.type == pygame.MOUSEMOTION:
                self.__gameMode.mouseMove(mouseX, mouseY, buttons)
        elif self.__mouseFocus:
            self.__mouseFocus = False
            self.__gameMode.mouseLeave()

    def processInput(self):
        if self.__gameMode is not None:
            self.__gameMode.processInput()
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.__running = False
                break
            elif event.type == pygame.KEYDOWN \
                    or event.type == pygame.KEYUP:
                if event.key == pygame.K_ESCAPE:
                    self.__running = False
                    break
                self.__processKeyEvent(event)
            elif event.type == pygame.ACTIVEEVENT:
                if event.state & pygame.APPFOCUSMOUSE == pygame.APPFOCUSMOUSE:
                    self.__processMouseEvent(event)
            elif event.type == pygame.MOUSEBUTTONDOWN \
                    or event.type == pygame.MOUSEBUTTONUP \
                    or event.type == pygame.MOUSEWHEEL \
                    or event.type == pygame.MOUSEMOTION:
                self.__processMouseEvent(event)

    def update(self):
        if self.__gameMode is not None:
            self.__gameMode.update()

    def render(self):
        # Render world in a surface
        renderSurface = Surface((self.__renderWidth, self.__renderHeight))
        if self.__gameMode is not None:
            self.__gameMode.render(renderSurface)

        # Scale rendering to window size
        windowWidth, windowHeight = self.__window.get_size()
        renderRatio = self.__renderWidth / self.__renderHeight
        windowRatio = windowWidth / windowHeight
        if windowRatio <= renderRatio:
            rescaledSurfaceWidth = windowWidth
            rescaledSurfaceHeight = int(windowWidth / renderRatio)
            self.__rescaledX = 0
            self.__rescaledY = (windowHeight - rescaledSurfaceHeight) // 2
        else:
            rescaledSurfaceWidth = int(windowHeight * renderRatio)
            rescaledSurfaceHeight = windowHeight
            self.__rescaledX = (windowWidth - rescaledSurfaceWidth) // 2
            self.__rescaledY = 0

        # Scale the rendering to the window/screen size
        rescaledSurface = pygame.transform.scale(
            renderSurface, (rescaledSurfaceWidth, rescaledSurfaceHeight)
        )
        self.__rescaledScaleX = rescaledSurface.get_width() / renderSurface.get_width()
        self.__rescaledScaleY = rescaledSurface.get_height() / renderSurface.get_height()
        self.__window.blit(rescaledSurface, (self.__rescaledX, self.__rescaledY))

    def run(self):
        # Main game loop
        while self.__running:
            self.processInput()
            self.update()
            self.render()

            pygame.display.update()
            self.__clock.tick(30)

    # Quit

    def quit(self):
        pygame.quit()

    # Event handling

