import random
from typing import Tuple, Optional

import pygame
from pygame.surface import Surface

from core.constants import LayerValue, LayerValueRanges
from core.state import World
from ..LayerComponent import LayerComponent
from ..MouseButtons import MouseButtons
from ..Theme import Theme
from .GameMode import GameMode


class EditGameMode(GameMode):

    def __init__(self, theme: Theme, world: World):
        super().__init__(theme)
        self.__world = world
        self.__mouseButtonDown = False
        self.__layers = [
            LayerComponent(theme, world.getLayer(name), name)
            for name in ["ground", "impassable", "objects"]
        ]
        self.__font = theme.getFont("default")

        # Brush
        self.__brushLayer = "impassable"

    # Game loop

    def processInput(self):
        pass

    def update(self):
        pass

    def render(self, surface: Surface):
        for layer in self.__layers:
            layer.render(surface)
        color = self.theme.getFontColor("default")
        textSurfaces = [
            self.__font.render(message, False, color)
            for message in [
                f"Current brush: {self.__brushLayer}",
                "Left click: add",
                "right click: remove",
                "F1: Select 'impassable' layer",
                "F2: Select 'objects' layer"
            ]
        ]
        y = surface.get_height()
        for textSurface in reversed(textSurfaces):
            y -= textSurface.get_height() + 1
            surface.blit(textSurface, (0, y))

    # Keyboard handling

    def keyDown(self, key: int):
        if key == pygame.K_F1:
            self.__brushLayer = "impassable"
        elif key == pygame.K_F2:
            self.__brushLayer = "objects"

    # Mouse handling

    def __computeCellCoordintates(self, mouseX: int, mouseY: int) -> Optional[Tuple[int, int]]:
        tileWidth, tileHeight = self.theme.getTileSize("ground")
        cellX = mouseX // tileWidth
        cellY = mouseY // tileHeight
        if not (0 <= cellX < self.__world.width) \
        or not (0 <= cellY < self.__world.height):
            return None
        return cellX, cellY

    def __updateCell(self, cellX: int, cellY: int, buttons: MouseButtons):
        layer = self.__world.getLayer(self.__brushLayer)
        if buttons.button1:
            if layer.getValue(cellX, cellY) != LayerValue.NONE:
                return
            minValue = LayerValueRanges[self.__brushLayer][0]
            maxValue = LayerValueRanges[self.__brushLayer][1]
            value = random.randint(minValue, maxValue - 1)
            layer.setValue(cellX, cellY, LayerValue(value))
        elif buttons.button3:
            layer.setValue(cellX, cellY, LayerValue.NONE)

    def mouseButtonDown(self, mouseX: int, mouseY: int, buttons: MouseButtons):
        coords = self.__computeCellCoordintates(mouseX, mouseY)
        if coords is None:
            return
        cellX, cellY = coords
        self.__mouseButtonDown = True
        self.__updateCell(cellX, cellY, buttons)

    def mouseMove(self, mouseX: int, mouseY: int, buttons: MouseButtons):
        if not self.__mouseButtonDown:
            return
        coords = self.__computeCellCoordintates(mouseX, mouseY)
        if coords is None:
            return
        cellX, cellY = coords
        self.__updateCell(cellX, cellY, buttons)

    def mouseButtonUp(self, mouseX: int, mouseY: int, buttons: MouseButtons):
        self.__mouseButtonDown = False

    def mouseEnter(self, mouseX: int, mouseY: int, buttons: MouseButtons):
        self.__mouseButtonDown = False

    def mouseLeave(self):
        self.__mouseButtonDown = False
